﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Threading;

namespace Sudoko
{
    class solve_sudoko
    {
        //محمد سعید زندی،حسین حسینی
        private int[,,] valdisp = new int[9, 9, 10];
        private int[,] val = new int[9, 9];
        public Thread oThread;
       
        public delegate void SolveDelegate(int progress,int[,] val);
        public delegate void ResultDelegate(Color c,string t);
        public delegate int SolveSpeed();

        SolveDelegate print_sudoko;
        ResultDelegate result_sudoko;
        SolveSpeed speed;
        /// <summary>
        /// متد سازنده کلاس
        /// </summary>
        public solve_sudoko(SolveDelegate print, ResultDelegate result, SolveSpeed speed, int[,] vals)
        {
            print_sudoko = print;
            result_sudoko = result;
            for (int i = 0; i < 9; i++)
                for (int j = 0; j < 9; j++)
                    val[i, j] = vals[i, j];
            this.speed = speed;
            
        }
        
        /// <summary>
        /// تابع اجرا کردن یک نخ جدید برای حل جدول به صورت مستقل از فرم
        /// </summary>
        public void run()
        {
            
                oThread = new Thread(new ThreadStart(SolveSudokuDispStart));
                oThread.Start();
            
        }
        /// <summary>
        /// تابع بسستن نخ به منظور خروج از برنامه در حین محاسبه
        /// </summary>
        public void exit()
        {
            if (oThread.IsAlive)
                oThread.Abort();
        }

        /// <summary>
        /// در جدول می گردد تا یک خانه خالی پیدا کند
        /// </summary>
        private bool CercaVuota(out int x, out int y)
        {
            for (x = 0, y = 0; x < 9; x++)
                for (y = 0; y < 9; y++)
                    if (val[x, y] == 0)
                        return true;
            return false;
        }

        /// <summary>
        /// در سطر و ستون جدول می گردد تا یک خانه خالی با کمترین حالت موجود برای قرار دادن در آن  پیدا کند
        /// </summary>
        private bool CercaVuotaDisp(out int x, out int y)
        {
            int min = 10, minx = 0, miny = 0;

            for (x = 0, y = 0; x < 9; x++)
                for (y = 0; y < 9; y++)
                    if (val[x, y] == 0 && min > valdisp[x, y, 0])
                    {
                        min = valdisp[x, y, 0];
                        minx = x; miny = y;
                    }
            if (min != 10)
            {
                x = minx;
                y = miny;
                return true;
            }
            return false;
        }

        /// <summary>
        ///برسی می کند مقدار در یافتی در سطر و ستون (و سلولی که آنها در آن متقاطع اند) 
        /// یک بار بیشتر نیامده باشد
        /// </summary>
        /// /// <param name="v"> Y  و شماره ستون X مقدار خانه با شماره سطر</param>
        /// <param name="x">شماره سطر</param>
        /// <param name="y">شماره ستون</param>
        /// <returns>false در صورت یافتن عدد تکراری </returns>
        private bool Compatibile(int v, int x, int y)
        {
            int i, j;

            //Controllo che sulla riga e sulla colonna non ci sia già lo stesso numero
            for (i = 0; i < 9; i++)
            {
                if (val[x, i] == v && i != y)
                    return false;

                if (val[i, y] == v && i != x)
                    return false;
            }
            //Controllo che nel riquadro non ci sia già lo stesso numero
            for (i = 0; i < 3; i++)
                for (j = 0; j < 3; j++)
                {
                    if (i == x && j == y)
                        continue;
                    if (val[x / 3 * 3 + i, y / 3 * 3 + j] == v)
                        return false;
                }
            //Controllo 

            return true;

        }
        //محمد سعید زندی،حسین حسینی
        /// <summary>
        /// الگریتم حل به روش بازگشتی
        /// </summary>
        /// <returns></returns>
        private bool SolveSudoku()
        {
            int vx, vy;

            if (!CercaVuota(out vx, out vy))
                return true;
            for (int i = 1; i < 10; i++)
                if (Compatibile(i, vx, vy))
                {
                    val[vx, vy] = i;
                    if (SolveSudoku())
                        return true;
                    val[vx, vy] = 0;
                }
            return false;
        }

        /// <summary>
        ///  شروع حل جدول سوداکو 
        /// </summary>
        private void SolveSudokuDispStart()
        {
            // بررسی حالات ممکن برای خانه ها با توجه به ورودی های موجود
            for (int i = 0; i < 9; i++)
                for (int j = 0; j < 9; j++)
                {
                    valdisp[i, j, 0] = 0;
                    for (int v = 1; v < 10; v++)
                    {
                        if (val[i, j] != 0)//موجود نبود V اگر در این خانه ورودی وجود نداشت و در حوزه دید این خانه عدد
                            valdisp[i, j, v] = 0;
                        else if (Compatibile(v, i, j))
                        {
                            valdisp[i, j, v] = 1;
                            valdisp[i, j, 0]++;
                        }
                        else
                            valdisp[i, j, v] = 0;
                    }
                }

            // برسی حوزه دید خانه ها برای یافتن عدد تکراری
            for (int i = 0; i < 9; i++)
                for (int j = 0; j < 9; j++)
                {
                    if (val[i, j] != 0)
                    {
                        int v = val[i, j];
                        val[i, j] = 0;
                        if (!Compatibile(v, i, j))
                        {
                           
                            result_sudoko(Color.Red, "    غیرقابل حل...  ");
                            val[i, j] = v;
                           
                            
                            return;
                        }
                        val[i, j] = v;
                    }
                }

            //حل به روش پویا
            if (SolveSudokuDisp())//اگر حل شد
            {
                result_sudoko(Color.Green, "   حل شد   ");
            }
            else
            {
                result_sudoko(Color.Red, "    راه حلی وجود ندارد   ");
                
            }

            //م.ح.م.د س.ع.ی.د ز.ن.د.ی،ح.س.ی.ن ح.س.ی.ن.ی
            return;
        }

        /// <summary>
        /// الگریتم حل به روش پویا
        /// </summary>
        private bool SolveSudokuDisp()
        {
            int vx, vy;
            int[] xdisp = new int[9], ydisp = new int[9];
            int[,] qdisp = new int[3, 3];

            if (!CercaVuotaDisp(out vx, out vy))
                return true;
            for (int v = 1; v < 10; v++)
            {
                if (valdisp[vx, vy, v] == 1)
                {

                    //را در آنها قرار داد ذخیره می کند وموقتا قابلیت گرفتن این عدد را از آنها می گیرد v تمام خانه هایی که در این سطر وستون می توانند مقدار
                    for (int i = 0; i < 9; i++)
                    {

                        xdisp[i] = valdisp[i, vy, v];

                        if (valdisp[i, vy, v] != 0)
                        {
                            valdisp[i, vy, v] = 0;
                            valdisp[i, vy, 0]--;
                        }
                        //=========================
                        ydisp[i] = valdisp[vx, i, v];

                        if (valdisp[vx, i, v] != 0)
                        {
                            valdisp[vx, i, v] = 0;
                            valdisp[vx, i, 0]--;
                        }
                    }

                    //را در آنها قرار داد  ذخیره می کند وموقتا قابلیت گرفتن این عدد را از خانه های سلول می گیرد v تمام خانه هایی که در این سلول می توانند مقدار
                    for (int i = 0; i < 3; i++)
                        for (int j = 0; j < 3; j++)
                        {
                            qdisp[i, j] = valdisp[vx / 3 * 3 + i, vy / 3 * 3 + j, v];
                            if (valdisp[vx / 3 * 3 + i, vy / 3 * 3 + j, v] != 0)
                            {
                                valdisp[vx / 3 * 3 + i, vy / 3 * 3 + j, v] = 0;
                                valdisp[vx / 3 * 3 + i, vy / 3 * 3 + j, 0]--;
                            }
                        }
                    //===============این مقدار راموقتا در جدول قرار می دهد و جدول جدید را حل می کند=============== 

                    val[vx, vy] = v;

                    // if (checkBox1.Checked == true)//اگر کاربر در خاست نمایش آرام داشته باشد نخ 0.1 ثانیه متوقف می شود 
                    Thread.Sleep(speed());
                    //WriteCell(v, vx, vy, Color.BlueViolet);
                    
                    print_sudoko(progressbar(),val);
                    if (SolveSudokuDisp())
                        return true;

                    //========================================================================================
                    //تفییرات موقتی به حالت سابق بر می گردند
                    val[vx, vy] = 0;

                    for (int i = 0; i < 3; i++)
                        for (int j = 0; j < 3; j++)
                        {
                            if (0 != qdisp[i, j])
                            {
                                valdisp[vx / 3 * 3 + i, vy / 3 * 3 + j, v] = qdisp[i, j];
                                valdisp[vx / 3 * 3 + i, vy / 3 * 3 + j, 0]++;
                            }
                        }

                    for (int i = 0; i < 9; i++)
                    {
                        if (0 != xdisp[i])
                        {
                            valdisp[i, vy, v] = xdisp[i];
                            valdisp[i, vy, 0]++;
                        }
                        if (0 != ydisp[i])
                        {
                            valdisp[vx, i, v] = ydisp[i];
                            valdisp[vx, i, 0]++;
                        }
                    }
                }
            }
            return false;
            /////م/ح/م/د س/ع/ی/د ز/ن/د/ی،ح/س/ی/ن ح/س/ی/ن/ی
        }
        /// <summary>
        /// تابع برای اجرای صحیح پروگوس بار
        /// </summary>        
        int progressbar()
        {
            int y = 0;
            foreach (int n in val)
                if (n != 0)
                    y++;
            float r = ((y / 81.0f) * 100);
            return (int)r;
        }


    }
}
